﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Linq.Expressions;
using GodHand.Infrastructure.LinqParser;
using System.ComponentModel;
using System.Collections;

namespace GodHand.Infrastructure.BlueBubble
{
    public class EntitySet<T> : IOrderedQueryable<T>
    {
        public DbContext Context { get; private set; }
        string TableName { get; set; }
        bool NoLock { get; set; }
        bool NoTracking { get; set; }

        DbQueryProvider MyProvider { get; set; }
        Expression MyExpression { get; set; }

        public EntitySet(string tableName, bool noLock, bool noTracking, DbContext context)
        {
            if (context == null) { throw new ArgumentNullException("context"); }
            InitEntitySet(tableName, noLock, noTracking, Expression.Constant(this), context);
        }

        public EntitySet(string tableName, bool noLock, bool noTracking, Expression expression, DbContext context)
        {
            if (context == null || expression == null) { throw new ArgumentNullException(); }
            if (!typeof(IQueryable<T>).IsAssignableFrom(expression.Type)) { throw new ArgumentOutOfRangeException("expression"); }
            InitEntitySet(tableName, noLock, noTracking, expression, context);
        }

        void InitEntitySet(string tableName, bool noLock, bool noTracking, Expression expression, DbContext context)
        {
            if (expression == null) { throw new ArgumentNullException(); }
            this.TableName = tableName;
            this.NoLock = noLock;
            this.NoTracking = noTracking;
            this.Context = context;
            this.MyExpression = expression;
            this.MyProvider = new DbQueryProvider(this.TableName, this.NoLock, this.NoTracking, context);
            this.MyProvider.OnExecuted += new EventHandler(MyProvider_OnExecuted);
        }

        public Expression Expression
        {
            get { return this.MyExpression; }
        }

        public Type ElementType
        {
            get { return typeof(T); }
        }

        public IQueryProvider Provider
        {
            get { return this.MyProvider; }
        }

        public IEnumerator<T> GetEnumerator()
        {
            var result = this.MyProvider.Execute(this.MyExpression);
            return (result as List<object>).Select(c => (T)c).GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return ((IEnumerable)this.MyProvider.Execute(this.MyExpression)).GetEnumerator();
        }

        public override string ToString()
        {
            if (this.MyExpression.NodeType == ExpressionType.Constant &&
                ((ConstantExpression)this.MyExpression).Value == this)
            {
                return "Query(" + typeof(T) + ")";
            }
            else
            {
                return this.MyExpression.ToString();
            }
        }

        void MyProvider_OnExecuted(object sender, EventArgs e)
        {
            TrackEntity(sender);
        }

        void TrackEntity(object result)
        {
            if (this.NoTracking) { return; }
            if (result is IEnumerable<T>)
            {
                var list = result as IEnumerable<T>;
                if (list.Count() == 0) { return; }
                if (!(list.First() is INotifyPropertyChanged)) { return; }
                list.ForEach(c =>
                {
                    var notifyT = c as INotifyPropertyChanged;
                    //if (notifyT == null) { throw new Exception("EntitySet enable Tracking,but Entity no imp interface of INotifyPropertyChanged"); }
                    notifyT.PropertyChanged += new PropertyChangedEventHandler(notifyT_PropertyChanged);
                });
            }
            else if (result is INotifyPropertyChanged)
            {
                var notifyT = result as INotifyPropertyChanged;
                notifyT.PropertyChanged += new PropertyChangedEventHandler(notifyT_PropertyChanged);
            }
        }

        void notifyT_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            var cmd = this.Context.Manager.EntityCommands.Where(c => c is UpdateCommand).Select(c => (UpdateCommand)c).Where(c => c.Key == sender).FirstOrDefault();
            if (cmd == null)
            {
                cmd = new UpdateCommand(sender, this.TableName);
                this.Context.Manager.EntityCommands.Add(cmd);
            }
            cmd.AddChangedProperty(e.PropertyName);
        }

        public void Add(T item, bool INSERT_IDENTITY = false)
        {
            if (item == null) { throw new ArgumentNullException(); }
            var cmd = new InsertCommand(item, this.TableName, INSERT_IDENTITY);
            this.Context.Manager.EntityCommands.Add(cmd);
        }

        public void Update(Expression<Func<T, bool>> exp, object obj)
        {
            if (exp == null && !this.Context.AllowUpdateWithNoExp) { throw new Exception("不允许无条件更新操作."); }
            this.Context.Manager.EntityCommands.Add(new UpdateCommand(typeof(T), exp, obj, this.TableName));
        }

        public void Remove(T item)
        {
            if (item == null) { throw new Exception("删除的对象不能为null."); }
            this.Context.Manager.EntityCommands.Add(new DeleteCommand(item, this.TableName));
        }

        public void Remove(Expression<Func<T, bool>> exp)
        {
            if (exp == null && !this.Context.AllowDeleteWithNoExp) { throw new Exception("不允许无条件删除操作."); }
            this.Context.Manager.EntityCommands.Add(new DeleteCommand(typeof(T), exp, this.TableName));
        }
    }
}
