﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.Linq.Expressions;
using GodHand.Infrastructure.Repository;
using GodHand.Infrastructure.LinqParser;

namespace GodHand.Infrastructure.BlueBubble
{
    public class UpdateCommand : EntityCommand
    {
        string TableName { get; set; }
        Type EntityType { get; set; }
        Expression WhereExpression { get; set; }
        object Data { get; set; }

        int UpdateType { get; set; }

        public object Key { get { return this.Entity; } }
        List<string> ChangedPropertys { get; set; }
        object Entity { get; set; }

        public UpdateCommand(object entity, string tableName)
        {
            if (entity == null) { throw new ArgumentNullException(); }
            this.Entity = entity;
            this.EntityType = entity.GetType();
            this.TableName = tableName;
            this.ChangedPropertys = new List<string>();
            this.UpdateType = 0;
        }

        public UpdateCommand(Type entityType, Expression whereExp, object data, string tableName)
        {
            if (entityType == null || data == null) { throw new ArgumentNullException(); }
            this.WhereExpression = whereExp;
            this.EntityType = entityType;
            this.Data = data;
            this.TableName = tableName;
            this.UpdateType = 1;
        }

        public void AddChangedProperty(string propertyName)
        {
            if (this.EntityType.GetMetaData().GetColumns().Any(c => c.PrimaryKey && c.Name == propertyName)) { throw new Exception("不能更改实体主键."); }
            if (!this.ChangedPropertys.Contains(propertyName))
            {
                this.ChangedPropertys.Add(propertyName);
            }
        }

        public override int Execute(IDbConnection conn, IDbTransaction tran)
        {
            switch (this.UpdateType)
            {
                case 0: { return this.ExecuteEntityUpdate(conn, tran); }
                case 1: { return this.ExecuteExpUpdate(conn, tran); }
                default: { throw new Exception("不正确的更新操作."); }
            }
        }

        int ExecuteEntityUpdate(IDbConnection conn, IDbTransaction tran)
        {
            var tableAttr = this.EntityType.GetMetaData();
            if (string.IsNullOrWhiteSpace(this.TableName)) { this.TableName = tableAttr.TableName; }

            var builder = new StringBuilder();
            DynamicParameters param = new DynamicParameters();
            var columnNames = new List<string>();
            this.ChangedPropertys.ForEach(c =>
            {
                var columnName = tableAttr.GetColumns().Where(cc => cc.Name == c).First().ColumnName;
                columnNames.Add(columnName);
                param.Add(columnName, EmitHelper.GetValue(c, this.Entity));
            });
            builder.AppendFormat("UPDATE [{0}] SET {1}", this.TableName, string.Join(",", columnNames.Select(c => string.Format("{0}=@{0}", c))));

            var keyColumns = tableAttr.GetColumns().Where(c => c.PrimaryKey).ToList();
            if (keyColumns.Count == 0) { throw new Exception("不支持没有主键的实体更新操作."); }
            keyColumns.ForEach(c =>
            {
                param.Add(c.ColumnName, EmitHelper.GetValue(c.Name, this.Entity));
            });
            builder.AppendFormat(" WHERE {0}", string.Join(",", keyColumns.Select(c => string.Format("{0}=@{0}", c.ColumnName))));
            return conn.Execute(builder.ToString(), param, tran);
        }

        int ExecuteExpUpdate(IDbConnection conn, IDbTransaction tran)
        {
            var tableAttr = this.EntityType.GetMetaData();
            if (string.IsNullOrWhiteSpace(this.TableName)) { this.TableName = tableAttr.TableName; }

            var builder = new StringBuilder();
            DynamicParameters param = new DynamicParameters();
            var columnNames = new List<string>();
            var utableAttr = this.Data.GetType().GetMetaData();
            utableAttr.GetColumns().ForEach(c =>
            {
                if (c.Identity) { return; }
                columnNames.Add(c.ColumnName);
                param.Add(c.ColumnName, EmitHelper.GetValue(c.Name, this.Data));
            });
            builder.AppendFormat("UPDATE [{0}] SET {1}", this.TableName, string.Join(",", columnNames.Select(c => string.Format("{0}=@{0}", c))));
            if (this.WhereExpression != null)
            {
                var parseContext = new ParserContext(this.EntityType);
                var whereVisitor = new WhereExpressionVisitor();
                whereVisitor.Parse(this.WhereExpression, parseContext);
                builder.AppendFormat(" WHERE {0}", whereVisitor.WhereBuilder.Replace(string.Format("[{0}]", parseContext.TableAlias), string.Format("[{0}]", this.TableName)));
                parseContext.DbParameters.ForEach(c => { param.Add(c.Key, c.Value); }); ;
            }
            return conn.Execute(builder.ToString(), param, tran);
        }
    }
}
