﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace GodHand.Infrastructure
{
    public static class ConvertExtension
    {
        private static Dictionary<Type, Func<object, object>> dict = new Dictionary<Type, Func<object, object>>();

        static ConvertExtension()
        {
            dict.Add(typeof(byte), WrapValueConvert(Convert.ToByte));
            dict.Add(typeof(short), WrapValueConvert(Convert.ToInt16));
            dict.Add(typeof(int), WrapValueConvert(Convert.ToInt32));
            dict.Add(typeof(long), WrapValueConvert(Convert.ToInt64));

            dict.Add(typeof(ushort), WrapValueConvert(Convert.ToUInt16));
            dict.Add(typeof(uint), WrapValueConvert(Convert.ToUInt32));
            dict.Add(typeof(ulong), WrapValueConvert(Convert.ToUInt64));

            dict.Add(typeof(int?), WrapValueConvert(Convert.ToInt32));
            dict.Add(typeof(double), WrapValueConvert(Convert.ToDouble));
            dict.Add(typeof(float), WrapValueConvert(Convert.ToSingle));
            dict.Add(typeof(Guid), f => new Guid(f.ToString()));
            dict.Add(typeof(string), Convert.ToString);
        }

        private static Func<object, object> WrapValueConvert<T>(Func<object, T> input) where T : struct
        {
            return i =>
            {
                if (i == null || i is DBNull) { return null; }
                return input(i);
            };
        }

        public static bool CanConvertTo(this object obj, Type targetType)
        {
            return dict.ContainsKey(targetType);
        }

        public static T To<T>(this object obj)
        {
            return (T)To(obj, typeof(T));
        }

        public static T To<T>(this object obj, T defaultValue)
        {
            try
            {
                if (obj == null) { return defaultValue; }
                return (T)To(obj, typeof(T));
            }
            catch
            {
                return defaultValue;
            }
        }

        public static bool TryConvertTo(this object obj, Type targetType, out object value)
        {
            value = null;
            if (obj != null)
            {
                try
                {
                    if (obj.GetType() == targetType || targetType.IsAssignableFrom(obj.GetType()))
                    {
                        value = obj;
                    }
                    else if (dict.ContainsKey(targetType))
                    {
                        value = dict[targetType](obj);
                    }
                    else if (targetType.IsEnum)
                    {
                        value = Enum.Parse(targetType, obj.ToString(), true);
                    }
                    else
                    {
                        value = Convert.ChangeType(obj, targetType);
                    }
                    return true;
                }
                catch
                {
                    return false;
                }
            }
            else
            {
                if (!targetType.IsValueType)
                {
                    value = null;
                    return true;
                }
                else
                {
                    return false;
                }
            }

        }

        public static object To(this object obj, Type targetType)
        {
            if (obj != null)
            {
                if (obj.GetType() == targetType || targetType.IsAssignableFrom(obj.GetType()))
                {
                    return obj;
                }
                else if (dict.ContainsKey(targetType))
                {
                    return dict[targetType](obj);
                }
                else
                {
                    try
                    {
                        return Convert.ChangeType(obj, targetType);
                    }
                    catch
                    {
                        throw new NotImplementedException("未实现到" + targetType.Name + "的转换");
                    }
                }
            }
            else
            {
                if ((targetType.IsGenericType && targetType.GetGenericTypeDefinition() == typeof(Nullable<>)) || !targetType.IsValueType)
                {
                    return null;
                }
                else
                {
                    throw new ArgumentNullException("obj", "不能将null转换为" + targetType.Name);
                }
            }
        }
    }
}
