﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Linq.Expressions;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;
using System.Data;
using System.Collections.Specialized;
using System.Reflection;
using System.Collections;
using System.Security.Cryptography;

namespace GodHand.Infrastructure
{
    public static class InfrastructureExtension
    {
        static readonly Regex datetimeReg = new Regex("\"\\\\/Date\\((-?[0-9]+)(?:[a-zA-Z]|(?:\\+|-)[0-9]{4})?\\)\\\\/\\\"", RegexOptions.Compiled);
        public static void Reset(this byte[] bytes, byte v)
        {
            for (var i = 0; i < bytes.Length; i++) { bytes[i] = 0; }
        }
        public static decimal SumOrDefault<TSource>(this IQueryable<TSource> source, Expression<Func<TSource, decimal>> selector)
        {
            if (source.Any())
            {
                return source.Sum(selector);
            }
            return 0;
        }
        public static decimal SumOrDefault<TSource>(this IQueryable<TSource> source, Expression<Func<TSource, decimal?>> selector)
        {
            if (source.Any())
            {
                var a = source.Sum(selector);
                return a.HasValue ? a.Value : 0m;
            }
            return 0m;
        }
        public static int SumOrDefault<TSource>(this IQueryable<TSource> source, Expression<Func<TSource, int?>> selector)
        {
            if (source.Any())
            {
                var a = source.Sum(selector);
                return a.HasValue ? a.Value : 0;

            }
            return 0;
        }
        public static int SumOrDefault<TSource>(this IQueryable<TSource> source, Expression<Func<TSource, int>> selector)
        {
            if (source.Any())
            {
                var a = source.Sum(selector);
                return a;
            }
            return 0;
        }
        public static T Clone<T>(this T obj)
        {
            if (obj == null) { return default(T); }
            BinaryFormatter formatter = new BinaryFormatter();
            using (MemoryStream memoryStream = new MemoryStream())
            {
                formatter.Serialize(memoryStream, obj);
                memoryStream.Position = 0;
                T newObject = (T)formatter.Deserialize(memoryStream);
                return newObject;
            }
        }
        public static Guid GetGuid(this NameValueCollection collection, string name)
        {
            return new Guid(collection[name]);
        }
        public static decimal GetDecimal(this NameValueCollection collection, string name)
        {
            return decimal.Parse(collection[name]);
        }
        public static int GetInt(this NameValueCollection collection, string name)
        {
            return int.Parse(collection[name]);
        }
        public static bool GetBool(this NameValueCollection collection, string name)
        {
            return bool.Parse(collection[name]);
        }
        public static DateTime GetDateTime(this NameValueCollection collection, string name)
        {
            return DateTime.Parse(collection[name]);
        }
        static Dictionary<string, List<PropertyInfo>> copyTCache = new Dictionary<string, List<PropertyInfo>>();
        static List<Type> copyNotNullPrimitiveTypes = new List<Type> { typeof(Int16), typeof(Int64), typeof(byte), typeof(char), typeof(bool), typeof(DateTime), typeof(int), typeof(double), typeof(float), typeof(decimal) };
        public static void Copy<T, U>(this T des, U source, bool ignoreAttribute = false, params string[] excludeProperty)
        {
            var cacthPropertys = new List<PropertyInfo>();
            var ttype = typeof(T);

            var ps = ttype.GetProperties(BindingFlags.Public | BindingFlags.Instance | BindingFlags.GetProperty | BindingFlags.SetProperty);
            if (ignoreAttribute)
            {
                cacthPropertys.AddRange(ps);
            }
            else
            {
                var tHasCopyAtt = false;
                if (ttype.GetCustomAttributes(typeof(CopyAttribute), true).Length > 0) { tHasCopyAtt = true; }

                foreach (var p in ps)
                {
                    if (tHasCopyAtt)
                    {
                        if (p.GetCustomAttributes(typeof(CopyIgnoreAttribute), true).Length == 0) { cacthPropertys.Add(p); }
                    }
                    else
                    {
                        if (p.GetCustomAttributes(typeof(CopyAttribute), true).Length > 0) { cacthPropertys.Add(p); }
                    }
                }
            }
            foreach (var p in cacthPropertys)
            {
                if (excludeProperty != null && excludeProperty.Contains(p.Name)) { continue; }
                var sourceValue = EmitHelper.GetValue(p.Name, source);
                var desValue = EmitHelper.GetValue(p.Name, des);
                //if (!p.PropertyType.IsClass && sourceValue == null) { continue; }
                if (sourceValue == desValue) { continue; }
                if (sourceValue == null && copyNotNullPrimitiveTypes.Contains(p.PropertyType))
                {
                    continue;
                }
                EmitHelper.SetValue(p.Name, des, new object[] { sourceValue });
            }
        }
        public static U CopyConvert<U>(this object des, bool ignoreAttribute = false) where U : class,new()
        {
            if (des == null) { return null; }
            var u = new U { };
            u.Copy(des, ignoreAttribute);
            return u;
        }
        public static List<U> CopyConvertList<U>(this IList list, bool ignoreAttribute = false) where U : class,new()
        {
            var listU = new List<U>();
            foreach (var c in list)
            {
                listU.Add(c.CopyConvert<U>(ignoreAttribute));
            }
            return listU;
        }
        public static void AddRange<T>(this SortedSet<T> source, IEnumerable<T> range)
        {
            foreach (var item in range)
            {
                source.Add(item);
            }
        }
        public static TValue Get<TKey, TValue>(this IDictionary<TKey, TValue> source, TKey key)
        {
            if (source.ContainsKey(key)) { return source[key]; }
            return default(TValue);
        }
        public static void Set<TKey, TValue>(this IDictionary<TKey, TValue> source, TKey key, TValue value)
        {
            if (source.ContainsKey(key))
            {
                source[key] = value;
            }
            else
            {
                source.Add(key, value);
            }
        }
        public static void ForEach<T>(this IEnumerable<T> source, Action<T> action)
        {
            if (source == null) { return; }
            foreach (var s in source) { action(s); }
        }
        public static void ForEach<TKey, TValue>(this IDictionary<TKey, TValue> source, Action<KeyValuePair<TKey, TValue>> action)
        {
            if (source == null) { return; }
            foreach (var s in source) { action(s); }
        }
        public static void ForEach<TKey, TValue>(this IDictionary<TKey, TValue> source, Action<KeyValuePair<TKey, TValue>, int> action)
        {
            if (source == null) { return; }
            var i = 0;
            foreach (var s in source) { action(s, i); i++; }
        }
        public static byte[] SerializeBinary(this object obj)
        {
            using (MemoryStream ms = new MemoryStream())
            {
                BinaryFormatter formatter = new BinaryFormatter();
                formatter.Serialize(ms, obj);
                return ms.ToArray();
            }
        }
        public static object DeserializeBinary(this byte[] Buffer)
        {
            using (MemoryStream ms = new MemoryStream(Buffer))
            {
                BinaryFormatter formatter = new BinaryFormatter();
                return formatter.Deserialize(ms);
            }
        }
        public static bool IsAnonymousType(this Type type) { return type.Name.StartsWith("<>f__AnonymousType"); }
        public static T ReadData<T>(this IDataReader dataReader) where T : class ,new()
        {
            var t = new T();
            for (var i = 0; i < dataReader.FieldCount; i++)
            {
                var pname = dataReader.GetName(i);
                var val = dataReader.GetValue(i);
                val = val == DBNull.Value ? null : val;
                EmitHelper.SetValue(pname, t, new object[] { val });
            }
            return t;
        }
        public static T TryReadData<T>(this IDataReader dataReader) where T : class ,new()
        {
            var t = new T();
            for (var i = 0; i < dataReader.FieldCount; i++)
            {
                var pname = dataReader.GetName(i);
                var val = dataReader.GetValue(i);
                val = val == DBNull.Value ? null : val;
                try
                {
                    EmitHelper.SetValue(pname, t, new object[] { val });
                }
                catch { }
            }
            return t;
        }
    }
}
