﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GodHand.Infrastructure;

namespace GodHand.Infrastructure.LinqParser
{
    public class ParserContext
    {
        int TableIndex { get; set; }
        public string TableAlias { get { return string.Format("t{0}", this.TableIndex); } }
        int ParamIndex { get; set; }
        public string ParamName { get { return string.Format("p{0}", this.ParamIndex); } }
        public string SQL { get; set; }
        public Type EntityType { get; set; }
        public bool HasSelectMethod { get; set; }
        public TableAttribute TableAttribute { get; private set; }
        public Dictionary<string, object> DbParameters { get; set; }
        public List<string> LastSelectColumns { get; set; }

        List<string> SelectColumns { get; set; }
        List<SelectTypeConstructor> SelectTypes { get; set; }
        Dictionary<string, string> OrderBys { get; set; }
        List<string> GroupBys { get; set; }
        Dictionary<string, string> ColumnAliasMap { get; set; }

        public ParserContext(Type entityType)
        {
            if (entityType == null) { throw new ArgumentNullException("entityType"); }
            this.EntityType = entityType;
            this.TableAttribute = this.EntityType.GetMetaData();
            this.DbParameters = new Dictionary<string, object>();
            this.ColumnAliasMap = new Dictionary<string, string>();
            this.SelectColumns = new List<string>();
            this.SelectTypes = new List<SelectTypeConstructor>();
            this.OrderBys = new Dictionary<string, string>();
            this.GroupBys = new List<string>();
            this.TableAttribute.GetColumns().ForEach(c =>
            {
                this.ColumnAliasMap.Add(c.Name, string.IsNullOrEmpty(c.Alias) ? c.Name : c.Alias);
            });
        }

        public void IncreaseTableIndex()
        {
            this.TableIndex++;
        }
        public void IncreaseParamIndex()
        {
            this.ParamIndex++;
        }
        public string GetColumnName(string columnAliasName)
        {
            var col = this.TableAttribute.GetColumns().FirstOrDefault(c => c.ColumnName == columnAliasName);
            if (col == null)
            {
                if (!this.ColumnAliasMap.ContainsKey(columnAliasName)) { throw new Exception("Alias don't map Column."); }
                return this.GetColumnName(this.ColumnAliasMap[columnAliasName]);
            }
            return col.ColumnName;
        }
        public void AddColumnMap(string aliasName, string columnName)
        {
            if (!this.ColumnAliasMap.ContainsKey(aliasName))
            {
                this.ColumnAliasMap.Add(aliasName, columnName);
            }
        }
        public void AddSelectColumnName(string name)
        {
            if (!this.SelectColumns.Contains(name))
            {
                this.SelectColumns.Add(name);
            }
        }
        public void AddSelectType(SelectTypeConstructor type)
        {
            if (!this.SelectTypes.Contains(type))
            {
                this.SelectTypes.Add(type);
            }
        }
        public void AddOrderBy(string name, string dir)
        {
            var o = this.OrderBys.Get(name);
            if (o == null)
            {
                this.OrderBys.Add(name, dir);
            }
        }
        public void AddGroupBy(string groupby)
        {
            if (!this.GroupBys.Contains(groupby))
            {
                this.GroupBys.Add(groupby);
            }
        }
        public List<string> DefaultAllColumns
        {
            get
            {
                return this.TableAttribute.GetColumns().Select(c => c.ColumnName).ToList();
            }
        }
        public Dictionary<string, string> DefaultOrderBys
        {
            get
            {
                var ob = this.TableAttribute.GetColumns().Where(c => c.PrimaryKey).Select(c => c.ColumnName).ToDictionary(c => c, c => "DESC");
                return ob;
            }
        }

        public List<string> GetSelectColumns() { return this.SelectColumns.ToList(); }
        public SelectTypeConstructor GetSelectType()
        {
            if (this.SelectTypes.Count == 0) { return new SelectTypeConstructor { SelectType = this.EntityType, ConstructorInfo = null }; }
            return this.SelectTypes.First();
        }
        public Dictionary<string, string> GetOrderBys() { return this.OrderBys.ToDictionary(c => c.Key, c => c.Value); }
        public List<string> GetGroupBys() { return this.GroupBys.ToList(); }
        public void ClearSelectColumns() { this.SelectColumns.Clear(); }
        public void ClearSelectTypes() { this.SelectTypes.Clear(); }
        public void ClearOrderBys() { this.OrderBys.Clear(); }
        public void ClearGroupBys() { this.GroupBys.Clear(); }
    }
}
