﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Linq.Expressions;
using System.Reflection;
using System.Data;
using System.Collections.ObjectModel;

namespace GodHand.Infrastructure.LinqParser
{
    public class WhereExpressionVisitor : AbstractExpressionVisitor
    {
        ParserContext Context { get; set; }
        public StringBuilder WhereBuilder;

        public void Parse(Expression exp, ParserContext context)
        {
            if (exp == null || context == null) { throw new ArgumentNullException(); }
            this.Context = context;
            this.WhereBuilder = new StringBuilder();
            //this.CurrQueryClosure.WhereSQLs.Add(WhereBuilder);
            this.WhereBuilder.Append("(");
            this.Visit(exp);
            this.WhereBuilder.Append(")");
        }

        protected override Expression VisitBinary(BinaryExpression b)
        {
            switch (b.NodeType)
            {
                case ExpressionType.AndAlso:
                case ExpressionType.OrElse:
                    {
                        this.WhereBuilder.Append("(");
                        base.Visit(b.Left);
                        this.WhereBuilder.Append(")");
                        if (b.NodeType == ExpressionType.AndAlso) { this.WhereBuilder.Append(" AND "); }
                        else if (b.NodeType == ExpressionType.OrElse) { this.WhereBuilder.Append(" OR "); }
                        this.WhereBuilder.Append("(");
                        if (b.Right is MemberExpression)
                        {
                            this.ParseOneMemberCondition(b.Right as MemberExpression);
                        }
                        else
                        {
                            base.Visit(b.Right);
                        }
                        this.WhereBuilder.Append(")");
                        return null;
                    }
                case ExpressionType.LessThan:
                case ExpressionType.LessThanOrEqual:
                case ExpressionType.GreaterThan:
                case ExpressionType.GreaterThanOrEqual:
                case ExpressionType.Equal:
                case ExpressionType.NotEqual:
                    {
                        this.ParseOneCondition(b.Left);
                        switch (b.NodeType)
                        {
                            case ExpressionType.LessThan:
                                { this.WhereBuilder.Append("<"); break; }
                            case ExpressionType.LessThanOrEqual:
                                { this.WhereBuilder.Append("<="); break; }
                            case ExpressionType.GreaterThan:
                                { this.WhereBuilder.Append(">"); break; }
                            case ExpressionType.GreaterThanOrEqual:
                                { this.WhereBuilder.Append(">="); break; }
                            case ExpressionType.Equal:
                                { this.WhereBuilder.Append("="); break; }
                            case ExpressionType.NotEqual:
                                { this.WhereBuilder.Append("<>"); break; }
                        }
                        this.ParseOneCondition(b.Right);
                        return null;
                    }
            }
            return base.VisitBinary(b);
        }
        void ParseOneCondition(Expression exp)
        {
            if (this.IsConstantExpression(exp))
            {
                this.ParseConstant(exp);
            }
            else if (exp is MemberExpression)
            {
                var mexp = exp as MemberExpression;
                if (mexp.Expression == null || mexp.Expression.NodeType == ExpressionType.Parameter)
                {
                    this.ParseEntityMember(exp as MemberExpression);
                }
                else if (mexp.Expression != null && (mexp.Expression.Type == typeof(DateTime) || mexp.Expression.Type == typeof(DateTime?)))
                {
                    this.ParseDateTimeProperty(mexp);
                }
                else if (mexp.Expression != null && mexp.Expression.Type == typeof(string))
                {
                    this.ParseStringProperty(mexp);
                }
                else
                { this.ParseOneCondition(mexp.Expression); }
            }
            else if (exp is UnaryExpression)
            {
                this.ParseOneCondition((exp as UnaryExpression).Operand);
            }
            else if (exp is LambdaExpression)
            {
                this.ParseOneCondition((exp as LambdaExpression).Body);
            }
            else if (exp is MethodCallExpression)
            {
                this.VisitMethodCall(exp as MethodCallExpression);
            }
            else { throw new Exception("未能解析条件表达式类型."); }
        }
        void ParseConstant(Expression cExp)
        {
            if (cExp == null) { return; }
            var paramName = this.CreateDbParamName();
            var value = GetConstant(cExp);
            if (value == null)
            {
                if (this.WhereBuilder.ToString().EndsWith("="))
                {
                    this.WhereBuilder.Remove(this.WhereBuilder.Length - 1, 1);
                    this.WhereBuilder.Append(" IS NULL ");
                    return;
                }
                else if (this.WhereBuilder.ToString().EndsWith("<>"))
                {
                    this.WhereBuilder.Remove(this.WhereBuilder.Length - 2, 2);
                    this.WhereBuilder.Append(" IS NOT NULL ");
                    return;
                }
            }
            this.WhereBuilder.AppendFormat("@{0}", paramName);
            this.Context.DbParameters.Set(paramName, value);
        }
        string CreateDbParamName()
        {
            var paramName = this.Context.ParamName;
            this.Context.IncreaseParamIndex();
            this.Context.DbParameters.Add(paramName, null);
            return paramName;
        }
        void ParseEntityMember(MemberExpression mExp)
        {
            if (mExp == null) { return; }
            var proName = mExp.Member.Name;
            var originalName = this.Context.GetColumnName(proName);
            this.Context.AddSelectColumnName(originalName);
            this.WhereBuilder.AppendFormat("[{0}].[{1}]", this.Context.TableAlias, originalName);
        }
        void ParseStringProperty(MemberExpression mexp)
        {
            var memberName = mexp.Member.Name.ToLower();
            switch (memberName)
            {
                case "length":
                    {
                        this.WhereBuilder.Append(" LEN(");
                        this.ParseEntityMember(mexp.Expression as MemberExpression);
                        this.WhereBuilder.Append(") ");
                        break;
                    }
                default: { throw new Exception(string.Format("未能解析System.String类型的属性{0}.", memberName)); }
            }
        }
        void ParseDateTimeProperty(MemberExpression mexp)
        {
            var memberName = mexp.Member.Name.ToLower();
            switch (memberName)
            {
                case "date":
                    {
                        this.WhereBuilder.Append(" CONVERT(");
                        this.ParseEntityMember(mexp.Expression as MemberExpression);
                        this.WhereBuilder.Append(",DATE) ");
                        break;
                    }
                case "year":
                    {
                        this.WhereBuilder.Append(" DATEPART(year,");
                        this.ParseEntityMember(mexp.Expression as MemberExpression);
                        this.WhereBuilder.Append(") ");
                        break;
                    }
                case "month":
                    {
                        this.WhereBuilder.Append(" DATEPART(month,");
                        this.ParseEntityMember(mexp.Expression as MemberExpression);
                        this.WhereBuilder.Append(") ");
                        break;
                    }
                case "day":
                    {
                        this.WhereBuilder.Append(" DATEPART(day,");
                        this.ParseEntityMember(mexp.Expression as MemberExpression);
                        this.WhereBuilder.Append(") ");
                        break;
                    }
                case "dayofyear":
                    {
                        this.WhereBuilder.Append(" DATEPART(dayofyear,");
                        this.ParseEntityMember(mexp.Expression as MemberExpression);
                        this.WhereBuilder.Append(") ");
                        break;
                    }
                case "dayofweek":
                    {
                        this.WhereBuilder.Append(" DATEPART(weekday,");
                        this.ParseEntityMember(mexp.Expression as MemberExpression);
                        this.WhereBuilder.Append(") ");
                        break;
                    }
                case "hour":
                    {
                        this.WhereBuilder.Append(" DATEPART(hour,");
                        this.ParseEntityMember(mexp.Expression as MemberExpression);
                        this.WhereBuilder.Append(") ");
                        break;
                    }
                case "minute":
                    {
                        this.WhereBuilder.Append(" DATEPART(minute,");
                        this.ParseEntityMember(mexp.Expression as MemberExpression);
                        this.WhereBuilder.Append(") ");
                        break;
                    }
                case "second":
                    {
                        this.WhereBuilder.Append(" DATEPART(second,");
                        this.ParseEntityMember(mexp.Expression as MemberExpression);
                        this.WhereBuilder.Append(") ");
                        break;
                    }
                case "millisecond":
                    {
                        this.WhereBuilder.Append(" DATEPART(millisecond,");
                        this.ParseEntityMember(mexp.Expression as MemberExpression);
                        this.WhereBuilder.Append(") ");
                        break;
                    }
                default: { throw new Exception(string.Format("未能解析DateTime的属性{0}.", memberName)); }
            }
        }
        bool IsConstantExpression(Expression exp)
        {
            if (exp is ConstantExpression) { return true; }
            if (exp is MemberExpression) { return IsConstantExpression((exp as MemberExpression).Expression); }
            if (exp is UnaryExpression) { return IsConstantExpression((exp as UnaryExpression).Operand); }
            return false;
        }
        object GetConstant(Expression exp, List<Tuple<string, MemberTypes>> pros = null)
        {
            if (pros == null) { pros = new List<Tuple<string, MemberTypes>>(); }
            if (exp is ConstantExpression)
            {
                var cExp = exp as ConstantExpression;
                if (pros.Count == 0) { return cExp.Value; }
                object value = cExp.Value;
                pros.ForEach(c =>
                {
                    switch (c.Item2)
                    {
                        case MemberTypes.Field: { value = value.GetType().GetField(c.Item1).GetValue(value); break; }
                        case MemberTypes.Property: { value = value.GetType().GetProperty(c.Item1).GetValue(value, null); break; }
                        default: { throw new Exception("未能识别的成员类型。"); }
                    }
                });
                return value;
            }
            else if (exp is MemberExpression)
            {
                var mExp = ((exp) as MemberExpression);
                var memberInfo = mExp.Member;
                if (!(memberInfo.DeclaringType.IsGenericType && memberInfo.DeclaringType.GetGenericTypeDefinition() == typeof(Nullable<>) && memberInfo.Name == "Value"))
                {
                    pros.Insert(0, Tuple.Create(memberInfo.Name, memberInfo.MemberType));
                }
                return GetConstant(mExp.Expression, pros);
            }
            else if (exp is UnaryExpression)
            {
                return GetConstant((exp as UnaryExpression).Operand, pros);
            }
            else
            {
                throw new Exception("未能解析常量表达式的值.");
            }
        }
        void ParseArrayContains(MemberExpression memberExp, Expression arrExp)
        {
            var arr = GetConstant(arrExp) as System.Collections.IEnumerable;
            if (arr == null || !arr.GetEnumerator().MoveNext()) { throw new Exception("未能解析数组Contains，数组为null或元素个数为0."); }
            if (memberExp == null) { throw new Exception("未能解析数组Contains中的MemberExpression."); }
            ParseEntityMember(memberExp);
            Type memberType = null;
            switch (memberExp.Member.MemberType)
            {
                case MemberTypes.Property: { memberType = (memberExp.Member as PropertyInfo).PropertyType; break; }
                case MemberTypes.Field: { memberType = (memberExp.Member as FieldInfo).FieldType; break; }
                default: { throw new Exception("未能识别的成员。"); }
            }
            var column = this.Context.TableAttribute.GetColumns().Where(c => c.Name == memberExp.Member.Name).First();
            var plist = new List<string>();

            foreach (var s in arr)
            {
                var paramName = CreateDbParamName();
                this.Context.DbParameters.Set(paramName, s);
                plist.Add(string.Format("@{0} ", paramName));
            }
            this.WhereBuilder.AppendFormat(" IN({0})", string.Join(",", plist));
        }
        protected override Expression VisitMethodCall(MethodCallExpression m)
        {
            ParseMethodCallExpression(m.Object, m.Method, m.Arguments);
            return null;
            //return base.VisitMethodCall(m);
        }
        void ParseOneMemberCondition(MemberExpression mExp)
        {
            if (mExp == null) { return; }
            var proName = mExp.Member.Name;
            this.WhereBuilder.AppendFormat("[{0}].[{1}]", this.Context.TableAlias, proName);
            var paramName = this.CreateDbParamName();
            if (mExp.Type == typeof(bool))
            {
                this.WhereBuilder.Append("=");
                this.WhereBuilder.AppendFormat("@{0}", paramName);
                this.Context.DbParameters.Set(paramName, true);
            }
            else if (mExp.Type == typeof(bool?))
            {
                this.WhereBuilder.Append("<>");
                this.WhereBuilder.AppendFormat("@{0}", paramName);
                this.Context.DbParameters.Set(paramName, false);
            }
        }
        void ParseMethodCallExpression(Expression obj, MethodInfo methodInfo, ReadOnlyCollection<Expression> args)
        {
            var methodName = methodInfo.Name.ToLower();
            if (methodName == "contains")
            {
                if (IsConstantExpression(obj) && args.Count == 1)
                {
                    ParseArrayContains(args[0] as MemberExpression, obj);
                    return;
                }
                if (IsConstantExpression(args[0]) && args.Count == 2)
                {
                    ParseArrayContains(args[1] as MemberExpression, args[0]);
                    return;
                }
            }
            if (obj is MethodCallExpression)
            {
                var m = obj as MethodCallExpression;
                ParseMethodCallExpression(m.Object, m.Method, m.Arguments);
                ParseMethodCall(null, methodInfo, args);
            }
            else if (obj is MemberExpression)
            {
                ParseMethodCall(obj, methodInfo, args);
            }
            else { throw new Exception("未能解析MethodCall表达式"); }
        }
        void ParseMethodCall(Expression obj, MethodInfo methodInfo, ReadOnlyCollection<Expression> args)
        {
            var typeName = methodInfo.DeclaringType.Name.ToLower();
            switch (typeName)
            {
                case "string": { ParseStringMethodCall(obj, methodInfo, args); break; }
                case "datetime": { ParseDateTimeMethodCall(obj, methodInfo, args); break; }
                default: { throw new Exception(string.Format("暂时不支持类型{0}的方法调用.", typeName)); }
            }
        }
        void ParseStringMethodCall(Expression obj, MethodInfo methodInfo, ReadOnlyCollection<Expression> args)
        {
            var methodName = methodInfo.Name.ToLower();
            switch (methodName)
            {
                case "contains":
                    {
                        this.ParseEntityMember(obj as MemberExpression);
                        this.WhereBuilder.Append(" LIKE '%'+");
                        foreach (var arg in args) { ParseOneCondition(arg); }
                        this.WhereBuilder.Append("+'%'");
                        break;
                    }
                case "startswith":
                    {
                        this.ParseEntityMember(obj as MemberExpression);
                        this.WhereBuilder.Append(" LIKE ");
                        foreach (var arg in args) { ParseOneCondition(arg); }
                        this.WhereBuilder.Append("+'%'");
                        break;
                    }
                case "endswith":
                    {
                        this.ParseEntityMember(obj as MemberExpression);
                        this.WhereBuilder.Append(" LIKE '%'+");
                        foreach (var arg in args) { ParseOneCondition(arg); }
                        break;
                    }
                case "trim":
                    {
                        if (args.Count == 0)
                        {
                            this.WhereBuilder.Append(" LTRIM(RTRIM(");
                            this.ParseEntityMember(obj as MemberExpression);
                            this.WhereBuilder.Append(" ))");
                        }
                        else { throw new Exception("Trim具有到 SQL 的转换，但重载“System.String Trim(Char[])”没有。"); }
                        break;
                    }
                case "trimstart":
                    {
                        if (args.Count == 0)
                        {
                            this.WhereBuilder.Append(" LTRIM(");
                            this.ParseEntityMember(obj as MemberExpression);
                            this.WhereBuilder.Append(" )");
                        }
                        else { throw new Exception("TrimStart具有到 SQL 的转换，但重载“System.String TrimStart(Char[])”没有。"); }
                        break;
                    }
                case "trimend":
                    {
                        if (args.Count == 0)
                        {
                            this.WhereBuilder.Append(" LTRIM(RTRIM(");
                            this.ParseEntityMember(obj as MemberExpression);
                            this.WhereBuilder.Append(" ))");
                        }
                        else { throw new Exception("TrimEnd具有到 SQL 的转换，但重载“System.String TrimEnd(Char[])”没有。"); }
                        break;
                    }
                case "replace":
                    {
                        if (args.Count == 2)
                        {
                            this.WhereBuilder.Append(" REPLACE(");
                            this.ParseEntityMember(obj as MemberExpression);
                            this.WhereBuilder.Append(",");
                            ParseOneCondition(args[0]);
                            this.WhereBuilder.Append(",");
                            ParseOneCondition(args[1]);
                            this.WhereBuilder.Append(")");
                        }
                        else { throw new Exception("Replace方法只有两个参数的方法。"); }
                        break;
                    }
                case "substring":
                    {
                        if (args.Count == 1 || args.Count == 2)
                        {
                            this.WhereBuilder.Append(" SUBSTRING(");
                            this.ParseEntityMember(obj as MemberExpression);
                            this.WhereBuilder.Append(",");
                            ParseOneCondition(args[0]);
                            this.WhereBuilder.Append("+1,");
                            if (args.Count == 1)
                            {
                                this.WhereBuilder.Append("LEN(");
                                this.ParseEntityMember(obj as MemberExpression);
                                this.WhereBuilder.Append(")");
                                this.WhereBuilder.Append("-");
                                ParseOneCondition(args[0]);
                                this.WhereBuilder.Append("+1");
                            }
                            else
                            {
                                ParseOneCondition(args[1]);
                            }
                            this.WhereBuilder.Append(")");
                        }
                        else { throw new Exception("Replace方法只有两个参数的方法。"); }
                        break;
                    }
                case "remove":
                    {
                        if (args.Count == 1 || args.Count == 2)
                        {
                            this.WhereBuilder.Append(" REPLACE(");
                            this.ParseEntityMember(obj as MemberExpression);
                            this.WhereBuilder.Append(",");

                            this.WhereBuilder.Append(" SUBSTRING(");
                            this.ParseEntityMember(obj as MemberExpression);
                            this.WhereBuilder.Append(",");
                            ParseOneCondition(args[0]);
                            this.WhereBuilder.Append("+1,");
                            if (args.Count == 1)
                            {
                                this.WhereBuilder.Append("LEN(");
                                this.ParseEntityMember(obj as MemberExpression);
                                this.WhereBuilder.Append(")");
                                this.WhereBuilder.Append("-");
                                ParseOneCondition(args[0]);
                                this.WhereBuilder.Append("+1");
                            }
                            else
                            {
                                ParseOneCondition(args[1]);
                            }
                            this.WhereBuilder.Append(")");

                            this.WhereBuilder.Append(",'')");
                        }
                        else { throw new Exception("Replace方法只有两个参数的方法。"); }
                        break;
                    }
                default: { throw new Exception(string.Format("未能解析方法", methodName)); }
            }
        }
        void ParseDateTimeMethodCall(Expression obj, MethodInfo methodInfo, ReadOnlyCollection<Expression> args)
        {
            var methodName = methodInfo.Name.ToLower();
            switch (methodName)
            {
                case "addyears":
                    {
                        this.WhereBuilder.Append(" DATEADD(year,");
                        foreach (var arg in args) { ParseOneCondition(arg); }
                        this.WhereBuilder.Append(",");
                        this.ParseEntityMember(obj as MemberExpression);
                        this.WhereBuilder.Append(")");
                        break;
                    }
                case "addmonths":
                    {
                        this.WhereBuilder.Append(" DATEADD(month,");
                        foreach (var arg in args) { ParseOneCondition(arg); }
                        this.WhereBuilder.Append(",");
                        this.ParseEntityMember(obj as MemberExpression);
                        this.WhereBuilder.Append(")");
                        break;
                    }
                case "adddays":
                    {
                        this.WhereBuilder.Append(" DATEADD(day,");
                        foreach (var arg in args) { ParseOneCondition(arg); }
                        this.WhereBuilder.Append(",");
                        this.ParseEntityMember(obj as MemberExpression);
                        this.WhereBuilder.Append(")");
                        break;
                    }
                case "addhours":
                    {
                        this.WhereBuilder.Append(" DATEADD(hour,");
                        foreach (var arg in args) { ParseOneCondition(arg); }
                        this.WhereBuilder.Append(",");
                        this.ParseEntityMember(obj as MemberExpression);
                        this.WhereBuilder.Append(")");
                        break;
                    }
                case "addminutes":
                    {
                        this.WhereBuilder.Append(" DATEADD(minute,");
                        foreach (var arg in args) { ParseOneCondition(arg); }
                        this.WhereBuilder.Append(",");
                        this.ParseEntityMember(obj as MemberExpression);
                        this.WhereBuilder.Append(")");
                        break;
                    }
                case "addseconds":
                    {
                        this.WhereBuilder.Append(" DATEADD(second,");
                        foreach (var arg in args) { ParseOneCondition(arg); }
                        this.WhereBuilder.Append(",");
                        this.ParseEntityMember(obj as MemberExpression);
                        this.WhereBuilder.Append(")");
                        break;
                    }
                case "addmilliseconds":
                    {
                        this.WhereBuilder.Append(" DATEADD(millisecond,");
                        foreach (var arg in args) { ParseOneCondition(arg); }
                        this.WhereBuilder.Append(",");
                        this.ParseEntityMember(obj as MemberExpression);
                        this.WhereBuilder.Append(")");
                        break;
                    }
                default: { throw new Exception(string.Format("未能解析方法", methodName)); }
            }
        }
        protected override Expression VisitUnary(UnaryExpression u)
        {
            switch (u.NodeType)
            {
                case ExpressionType.Not:
                    {
                        this.WhereBuilder.Append(" not ");
                        break;
                    }
            }
            this.ParseOneMemberCondition(u.Operand as MemberExpression);
            return base.VisitUnary(u);
        }
    }
}
