﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.Reflection;
using System.Linq.Expressions;
using System.Threading;
using GodHand.Infrastructure.LinqParser;
using GodHand.Infrastructure.BlueBubble;
using System.Data.Common;

namespace GodHand.Infrastructure.Repository
{
    public static class DbExtension
    {
        public static DbEntityTable<T> GetTable<T>(this IDbConnection db) where T : class
        {
            return new DbEntityTable<T>(db, null, false);
        }
        public static DbEntityTable<T> GetTable<T>(this IDbConnection db, string tableName = null) where T : class
        {
            return new DbEntityTable<T>(db, tableName, false);
        }
        public static DbEntityTable<T> GetTable<T>(this IDbConnection db, bool noLock = false) where T : class
        {
            return new DbEntityTable<T>(db, null, noLock);
        }
        public static DbEntityTable<T> GetTable<T>(this IDbConnection db, string tableName = null, bool noLock = false) where T : class
        {
            if (typeof(T).GetMetaData().GetReferences().Count > 0) { throw new Exception("DbEntityTable暂时只支持单表."); }
            return new DbEntityTable<T>(db, tableName, noLock);
        }
        public static void UsingTransaction(this IDbConnection db, Action<IDbTransaction> action)
        {
            if (db == null || action == null) { throw new ArgumentNullException(); }
            if (db.State != ConnectionState.Closed) { db.Close(); }
            db.Open();
            using (var tran = db.BeginTransaction())
            {
                try
                {
                    action(tran);
                    tran.Commit();
                }
                catch (Exception)
                {
                    try { tran.Rollback(); }
                    catch { }
                    throw;
                }
            }
        }
    }

    public class DbEntityTable<T> where T : class
    {
        public IDbConnection DbConn { get; private set; }
        public string TableName { get; set; }
        public bool NoLock { get; set; }

        Type EntityType { get; set; }
        QueryExpressionClosure QEC { get; set; }
        Criteria Criteria { get; set; }
        public DbEntityTable(IDbConnection conn, string tableName, bool noLock, Type entityType = null)
        {
            this.DbConn = conn;
            this.TableName = tableName;
            this.NoLock = noLock;
            this.QEC = new QueryExpressionClosure();
            this.EntityType = entityType != null ? entityType : typeof(T);
        }

        static void DeadLockRetry(int time, Action action)
        {
            while (time > 0)
            {
                try
                {
                    action();
                    break;
                }
                catch (DbException sex)
                {
                    if (time > 1 && sex.Message.Contains("was deadlocked on lock"))
                    {
                        time--;
                        Thread.Sleep(500);
                    }
                    else
                    {
                        throw;
                    }
                }
            }
        }

        static void ParseCriteria(IList<CriteriaPre> whereCriteriaPres, QueryCommnClosure qcc, ParserContext context)
        {
            if (whereCriteriaPres == null || whereCriteriaPres.Count == 0) { return; }
            var preSql = string.Join(" and ", whereCriteriaPres.Select(c =>
            {

                var columnName = context.GetColumnName(c.ColumnName);
                var paramName = context.ParamName;
                if (c.Operator != "null" && c.Operator != "notnull")
                {
                    context.DbParameters.Add(context.ParamName, c.Value);
                    context.IncreaseParamIndex();
                }
                switch (c.Operator)
                {
                    case "=": { return string.Format("[{0}]=@{1}", columnName, paramName); }
                    case "<>":
                    case "!=": { return string.Format("[{0}]<>@{1}", columnName, paramName); }
                    case "like": { if (c.ColumnType != "string") { throw new Exception("like操作符只支持string类型！"); } return string.Format("[{0}] LIKE '%'+@{1}+'%'", columnName, paramName); }
                    case "%like": { if (c.ColumnType != "string") { throw new Exception("%like操作符只支持string类型！"); } return string.Format("[{0}] LIKE '%'+@{1}", columnName, paramName); }
                    case "like%": { if (c.ColumnType != "string") { throw new Exception("like%操作符只支持string类型！"); } return string.Format("[{0}] LIKE @{1}+'%'", columnName, paramName); }
                    case ">=":
                    case "<=":
                    case "<":
                    case ">":
                        {
                            switch (c.ColumnType)
                            {
                                case "int":
                                case "float":
                                case "date":
                                case "decimal": { return string.Format("[{0}]{1}@{2}", columnName, c.Operator, paramName); }
                                default: { throw new Exception(string.Format(">操作符不支持{0}类型", c.ColumnType)); }
                            }
                        }
                    case "null": { return string.Format("[{0}] IS NULL", columnName); }
                    case "notnull": { return string.Format("[{0}] IS NOT NULL", columnName); }
                    default: { throw new Exception(string.Format("不支持的操作符{0}", c.Operator)); }
                }
            }));
            qcc.WhereSQLs.Add(preSql);
        }

        static ParserContext GenQuerySQL(DbEntityTable<T> entiryTable)
        {
            var context = new ParserContext(entiryTable.EntityType);
            context.HasSelectMethod = entiryTable.QEC.QuerySelectExpressions.Count > 0;
            var qcc = entiryTable.QEC.PrepareContext(context);
            if (!string.IsNullOrWhiteSpace(entiryTable.TableName)) { qcc.TableName = entiryTable.TableName; }
            qcc.NoLock = entiryTable.NoLock;
            if (entiryTable.Criteria != null)
            {
                qcc.Take = entiryTable.Criteria.Limit;
                qcc.Skip = entiryTable.Criteria.Start;
                ParseCriteria(entiryTable.Criteria.Precidate, qcc, context);
                entiryTable.Criteria.OrderBy.ForEach(c => { qcc.OrderBys.Add(c.Key, c.Value); });
            }
            qcc.Generate(context);
            return context;
        }

        DbEntityTable<P> CloneMe<P>() where P : class
        {
            var me = new DbEntityTable<P>(this.DbConn, this.TableName, this.NoLock, this.EntityType);
            me.QEC = this.QEC.Clone();
            me.Criteria = this.Criteria.Clone();
            return me;
        }

        public DbEntityTable<T> Take(int take)
        {
            var me = this.CloneMe<T>();
            me.QEC.Take = take;
            return me;
        }
        public DbEntityTable<T> Skip(int skip)
        {
            var me = this.CloneMe<T>();
            me.QEC.Skip = skip;
            return me;
        }
        public DbEntityTable<T> Where(Expression<Func<T, bool>> whereExp)
        {
            if (whereExp == null) { return this; }
            var me = this.CloneMe<T>();
            me.QEC.WhereExpressions.Add(whereExp);
            return me;
        }
        public DbEntityTable<U> Select<U>(Expression<Func<T, U>> selectExp) where U : class
        {
            if (selectExp == null) { throw new ArgumentNullException(); }
            var me = this.CloneMe<U>();
            me.QEC.QuerySelectExpressions.Add(selectExp);
            return me;
        }
        public DbEntityTable<T> OrderBy(Expression<Func<T, object>> orderbyExp)
        {
            if (orderbyExp == null) { throw new ArgumentNullException(); }
            var me = this.CloneMe<T>();
            me.QEC.OrderByExpressions.Add(orderbyExp);
            return me;
        }
        public DbEntityTable<T> OrderByDescending(Expression<Func<T, object>> orderbyExp)
        {
            if (orderbyExp == null) { throw new ArgumentNullException(); }
            var me = this.CloneMe<T>();
            me.QEC.OrderByExpressions.Add(orderbyExp);
            return me;
        }
        public T First(Expression<Func<T, bool>> whereExp = null, IDbTransaction tran = null)
        {
            var t = FirstOrDefault(whereExp, tran);
            if (t == null) { throw new Exception("result of DbExtension's first invoke is null."); }
            return t;
        }
        public T FirstOrDefault(Expression<Func<T, bool>> whereExp = null, IDbTransaction tran = null)
        {
            var me = this.Where(whereExp);
            me.QEC.MethodName = "first";
            var context = GenQuerySQL(me);
            var result = me.DbConn.Query(context.SQL, context.DbParameters, tran);
            var cresult = context.GetSelectType().ConvertData(result, context);
            if (result.Count() > 0) { return cresult.First() as T; }
            return null;
        }
        public T[] ToArray(Expression<Func<T, bool>> whereExp = null, IDbTransaction tran = null)
        {
            return ToList(whereExp, tran).ToArray();
        }
        public List<T> ToList(Expression<Func<T, bool>> whereExp = null, IDbTransaction tran = null)
        {
            var me = this.Where(whereExp);
            me.QEC.MethodName = "tolist";
            var context = GenQuerySQL(me);
            var result = me.DbConn.Query(context.SQL, context.DbParameters, tran);
            var cresult = context.GetSelectType().ConvertData(result, context);
            return cresult.Select(c => (T)c).ToList();
        }
        public int Count(Expression<Func<T, bool>> whereExp = null, IDbTransaction tran = null)
        {
            var me = this.Where(whereExp);
            me.QEC.MethodName = "count";
            var context = GenQuerySQL(me);
            var result = me.DbConn.Query<int>(context.SQL, context.DbParameters, tran);
            if (result.Count() > 0) { return result.First(); }
            return 0;
        }
        public long LongCount(Expression<Func<T, bool>> whereExp = null, IDbTransaction tran = null)
        {
            var me = this.Where(whereExp);
            me.QEC.MethodName = "count";
            var context = GenQuerySQL(me);
            var result = me.DbConn.Query<long>(context.SQL, context.DbParameters, tran);
            if (result.Count() > 0) { return result.First(); }
            return 0;
        }
        static List<Type> EvalTypes = new List<Type> { typeof(ushort), typeof(ushort?), typeof(short), typeof(short?), typeof(uint), typeof(uint?), typeof(int), typeof(int?), typeof(ulong), typeof(ulong?), typeof(long), typeof(long?), typeof(float), typeof(float?), typeof(double), typeof(double?), typeof(decimal), typeof(decimal?) };
        dynamic InternalSum<U>(Expression<Func<T, U>> evalNumericExp, IDbTransaction tran = null)
        {
            var utype = typeof(U);
            if (!EvalTypes.Contains(utype)) { throw new Exception(string.Format("求值表达式不支持类型", utype.Name)); }
            if (evalNumericExp == null) { throw new Exception("表达式不能为空."); }
            var me = this.CloneMe<T>();
            me.QEC.EvalNumericExpression = evalNumericExp;
            me.QEC.MethodName = "sum";
            var context = GenQuerySQL(me);
            var result = me.DbConn.Query(context.SQL, context.DbParameters, tran);
            var v = (result.First() as IDictionary<string, object>).Values.First();
            return v == DBNull.Value ? null : v;
        }
        public U Sum<U>(Expression<Func<T, U>> evalNumericExp, IDbTransaction tran = null)
        {
            var result = InternalSum(evalNumericExp, tran);
            if (result != null) { return result; }
            throw new Exception("value of sum is null.");
        }
        public U SumOrDefault<U>(Expression<Func<T, U>> evalNumericExp, IDbTransaction tran = null)
        {
            var result = InternalSum(evalNumericExp, tran);
            if (result != null) { return result; }
            return default(U);
        }
        dynamic InternalMax<U>(Expression<Func<T, U>> evalNumericExp, IDbTransaction tran = null)
        {
            var utype = typeof(U);
            if (!EvalTypes.Contains(utype)) { throw new Exception(string.Format("求值表达式不支持类型", utype.Name)); }
            if (evalNumericExp == null) { throw new Exception("表达式不能为空."); }
            var me = this.CloneMe<T>();
            me.QEC.EvalNumericExpression = evalNumericExp;
            me.QEC.MethodName = "max";
            var context = GenQuerySQL(me);
            var result = me.DbConn.Query(context.SQL, context.DbParameters, tran);
            var v = (result.First() as IDictionary<string, object>).Values.First();
            return v == DBNull.Value ? null : v;
        }
        public U Max<U>(Expression<Func<T, U>> evalNumericExp, IDbTransaction tran = null)
        {
            var result = InternalMax(evalNumericExp, tran);
            if (result != null) { return result; }
            throw new Exception("value of max is null.");
        }
        public U MaxOrDefault<U>(Expression<Func<T, U>> evalNumericExp, IDbTransaction tran = null)
        {
            var result = InternalMax(evalNumericExp, tran);
            if (result != null) { return result; }
            return default(U);
        }
        dynamic InternalMin<U>(Expression<Func<T, U>> evalNumericExp, IDbTransaction tran = null)
        {
            var utype = typeof(U);
            if (!EvalTypes.Contains(utype)) { throw new Exception(string.Format("求值表达式不支持类型", utype.Name)); }
            if (evalNumericExp == null) { throw new Exception("表达式不能为空."); }
            var me = this.CloneMe<T>();
            me.QEC.EvalNumericExpression = evalNumericExp;
            me.QEC.MethodName = "min";
            var context = GenQuerySQL(me);
            var result = me.DbConn.Query(context.SQL, context.DbParameters, tran);
            var v = (result.First() as IDictionary<string, object>).Values.First();
            return v == DBNull.Value ? null : v;
        }
        public U Min<U>(Expression<Func<T, U>> evalNumericExp, IDbTransaction tran = null)
        {
            var result = InternalMin(evalNumericExp, tran);
            if (result != null) { return result; }
            throw new Exception("value of min is null.");
        }
        public U MinOrDefault<U>(Expression<Func<T, U>> evalNumericExp = null, IDbTransaction tran = null)
        {
            var result = InternalMin(evalNumericExp, tran);
            if (result != null) { return result; }
            return default(U);
        }
        dynamic InternalAverage<U>(Expression<Func<T, U>> evalNumericExp, IDbTransaction tran = null)
        {
            var utype = typeof(U);
            if (!EvalTypes.Contains(utype)) { throw new Exception(string.Format("求值表达式不支持类型", utype.Name)); }
            if (evalNumericExp == null) { throw new Exception("表达式不能为空."); }
            var me = this.CloneMe<T>();
            me.QEC.EvalNumericExpression = evalNumericExp;
            me.QEC.MethodName = "average";
            var context = GenQuerySQL(me);
            var result = me.DbConn.Query(context.SQL, context.DbParameters, tran);
            var v = (result.First() as IDictionary<string, object>).Values.First();
            return v == DBNull.Value ? null : v;
        }
        public U Average<U>(Expression<Func<T, U>> evalNumericExp, IDbTransaction tran = null)
        {
            var result = InternalAverage(evalNumericExp, tran);
            if (result != null) { return result; }
            throw new Exception("value of average is null.");
        }
        public U AverageOrDefault<U>(Expression<Func<T, U>> evalNumericExp, IDbTransaction tran = null)
        {
            var result = InternalAverage(evalNumericExp, tran);
            if (result != null) { return result; }
            return default(U);
        }
        public bool Any(Expression<Func<T, bool>> whereExp = null, IDbTransaction tran = null)
        {
            var me = this.Where(whereExp);
            me.QEC.MethodName = "any";
            var context = GenQuerySQL(me);
            var result = me.DbConn.Query<int>(context.SQL, context.DbParameters, tran);
            return result.First().Equals(1);
        }
        public bool Contains(Expression<Func<T, bool>> whereExp = null, IDbTransaction tran = null)
        {
            return Any(whereExp, tran);
        }
        public PageResult<T> Page(Criteria criteria, IDbTransaction tran = null)
        {
            var pageResult = new PageResult<T>();
            this.Criteria = criteria;
            var take = this.Criteria.Limit;
            var skip = this.Criteria.Start;
            this.Criteria.Limit = 0;
            this.Criteria.Start = 0;
            pageResult.TotalCount = Count(null, tran);
            if (pageResult.TotalCount == 0) { return pageResult; }

            this.Criteria.Limit = take;
            this.Criteria.Start = skip;
            pageResult.List = ToList(null, tran);
            return pageResult;
        }
        public void Insert(T data, IDbTransaction tran = null) { Insert(data, false, tran); }
        public void Insert(T data, bool IDENTITY_INSERT, IDbTransaction tran = null)
        {
            var cmd = new InsertCommand(data, this.TableName, IDENTITY_INSERT);
            cmd.Execute(this.DbConn, tran);
        }
        public void Update(object data, Expression<Func<T, bool>> whereExp = null, IDbTransaction tran = null)
        {
            if (whereExp == null) { throw new Exception("暂时不允许无条件的更新."); }
            var cmd=new UpdateCommand(this.EntityType, whereExp, data, this.TableName);
            cmd.Execute(this.DbConn, tran);
        }
        public void Delete(Expression<Func<T, bool>> whereExp = null, IDbTransaction tran = null)
        {
            if (whereExp == null) { throw new Exception("暂时不允许无条件的删除."); }
            var cmd = new DeleteCommand(this.EntityType, whereExp, this.TableName);
            cmd.Execute(this.DbConn, tran);
        }
    }
}
